from aiohttp import web
import json


class HTTPServer:
    def start_server(self, loop, interface, port, tweets):
        '''
        Starts the aiohttp server on the supplied interface and port.

        Arguments:
            loop: the asyncio event loop.
            interface (string): the interface to listen on.
            port (int): the port to listen on.
            tweets (list): the list of tweets to return.
        '''

        runner = self.___http_server(tweets)
        loop.run_until_complete(runner.setup())
        server = web.TCPSite(runner, interface, int(port))
        loop.run_until_complete(server.start())

    def ___http_server(self, tweets):
        '''
        Establishes the endponts and HTTP server configuration.

        Arguments:
            tweets (list): the list of tweets to return.

        Returns:
            runner: the app runner for asyncio.
        '''

        async def http_tweets(request):
            tweets_json = list(
                {
                    'created_at': x[0],
                    'full_text': x[1],
                    'id': x[2]
                }
                for x in tweets
            )
            return web.Response(
                body=json.dumps({'tweets': tweets_json}),
                headers={
                    'content-type': 'application/json'
                }
            )
        app = web.Application()
        app.router.add_get('/tweets', http_tweets)
        runner = web.AppRunner(app)
        return runner



from colored import fg, attr
from twittertail.exceptions import FailedToGetTweetsException
import asyncio
import sys


class RenderTweets:
    def get_tweets(
        self,
        tweets_response,
        count=None,
        refresh_tokens=False,
        last_id=None
    ):
        '''
        Gets the tweets from the GetTweetsAPI() instance.

        Arguments:
            tweets_response: the GetTweetsAPI() instance.
            count (int): the amount of tweets to return.
            refresh_tokens (bool): get new tokens before checking for tweets.
            last_id (int): the id of the latest seen tweet.

        Return:
            tweets_api (list): the tweets retrieved from GetTweetsAPI().
        '''

        tweets_api = list()
        try:
            tweets_api = tweets_response.get_tweets(
                count=count,
                refresh_tokens=refresh_tokens,
                last_id=last_id
            )
        except FailedToGetTweetsException as e:
            print(e)
            exit(1)
        return tweets_api

    def render_tweets(self, tweets):
        '''
        Prints out the tweets to stdout.

        Arguments:
            tweets (list): the list of tweet tuples to print.
        '''

        for tweet in tweets:
            # remove newlines in each tweet, so newlines differentiate between
            # tweets in the output
            sys.stdout.write(
                '[%s %s %s] %s\n'
                % (fg(111), tweet[0], attr(0), ' '.join(tweet[1].splitlines()))
            )
            sys.stdout.flush()

    async def render_5_most_recent_tweets(self, tweets, tweets_response):
        '''
        Gets the 5 most recent tweets and renders them.

        Arguments:
            tweets (list): the tweets to render.
            tweets_response: the GetTweetsAPI() instance.
        '''

        tweets += self.get_tweets(tweets_response, count=5)
        self.render_tweets(tweets)

    async def render_new_tweets_every_10_minutes(
        self,
        tweets,
        tweets_response
    ):
        '''
        Sets a loop to retrieve unseen tweets, checking every 10 minutes.

        Arguments:
            tweets (list): the tweets to render.
            tweets_response: the GetTweetsAPI() instance.
        '''

        while True:
            await asyncio.sleep(10 * 60)
            last_id = int(tweets[-1][2])
            new_tweets = self.get_tweets(
                tweets_response,
                count=None,
                refresh_tokens=True,
                last_id=last_id
            )
            self.render_tweets(new_tweets)
            if new_tweets:
                tweets += new_tweets



from twittertail.exceptions import (
    FailedToGetTwitterValueException,
    FailedToGetMainJsException,
    FailedToGetTweetsException
)
from twittertail.render_tweets import RenderTweets
from twittertail.get_tweets_api import GetTweetsAPI
from twittertail.get_twitter_values import GetTwitterValues
from twittertail.http_server import HTTPServer



from argparse import ArgumentParser
from .get_tweets_api import GetTweetsAPI
from .render_tweets import RenderTweets
from .http_server import HTTPServer
from .exceptions import FailedToGetTweetsException
import asyncio

app_description = '''
Monitors a Twitter account for tweet activity. On execution, the 5 most recent
tweets are displayed, and every 10 minutes new activity is displayed.
'''


def get_parser():
    parser = ArgumentParser(description=app_description)
    parser.add_argument(
        '-u',
        '--user',
        required=True,
        help='The value of the Twitter username to retrieve tweets from.'
    )
    # it's unclear whether retweets should be returned from the requirements
    # given, so the -r argument allows you to enable RT's (off by default).
    parser.add_argument(
        '-r',
        '--retweets',
        action='store_true',
        help='Show retweets.'
    )
    # the 'simple curl command' friendly API requirement is enabled by this
    # argument, which establishes an aiohttp server when enabled (off by
    # default). The requirements didn't specify HTTP, and curl has access
    # to a lot of protocols in most its standard OS compilations such as
    # file:// or a unix socket like the Docker daemon API, but HTTP was chosen
    # for its 'most obvious choice' properties.
    parser.add_argument(
        '-w',
        '--web-server',
        action='store_true',
        help='Run a simple HTTP server to get the tweets collected so far.'
    )
    parser.add_argument(
        '-i',
        '--interface',
        default='0.0.0.0',
        help='The interface to run the HTTP server on. Defaults to 0.0.0.0.'
    )
    parser.add_argument(
        '-p',
        '--port',
        default=9000,
        help='The port to run the HTTP server on. Defaults to 9000.'
    )
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()

    # the asyncio event loop
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    tweets = list()
    try:
        tweets_response = GetTweetsAPI(
            user=args.user,
            retweets=args.retweets
        )
    except FailedToGetTweetsException as e:
        print(e)
        exit(1)
    renderer = RenderTweets()

    # if the HTTP API is enabled, start an async web server
    if args.web_server:
        http_server = HTTPServer()
        http_server.start_server(loop, args.interface, args.port, tweets)

    task = loop.create_task(
        renderer.render_5_most_recent_tweets(tweets, tweets_response)
    )
    loop.run_until_complete(task)
    task = loop.create_task(
        renderer.render_new_tweets_every_10_minutes(
            tweets,
            tweets_response
        )
    )

    try:
        loop.run_forever()
    except (KeyboardInterrupt, SystemExit):
        loop.close()
        exit(0)



from twittertail.get_twitter_values import GetTwitterValues
from twittertail.exceptions import (
    FailedToGetTwitterValueException,
    FailedToGetTweetsException
)
from html import unescape
import requests
import re


class GetTweetsAPI:
    '''
    GetTweetsAPI mimmicks the process a browser uses when accessing a Twitter
    user's tweet timeline in an unauthenticated session. To do this, it:

    1. Gets the "guest token" from the Twitter markup, and uses it in the
    "x-guest-token" request header in the API call.

    2. Gets the bearer token from the Twitter main.js, and uses it in the
    "authorization" request header in the API call.

    3. Gets the user id for the supplied username from a GraphQL query.

    4. Queries the Twitter API at /2/timeline/profile/X.json, where X is the
    user id.

    Arguments:
        user (str): the username of the Twitter user to query against.
        retweets (bool): toggles whether to return retweets

    Raises:
        FailedToGetTweetsException: if the username is invalid by Twitter
        standards.
    '''

    def __init__(self, user, retweets=False):
        if len(user) > 15 or re.search('[^a-zA-Z0-9_]+', user):
            raise FailedToGetTweetsException(
                'Invalid username - Twitter usernames must be 15 or fewer '
                'characters in length, and must be alphanumeric only '
                '(with underscores).'
            )
        self.user = user
        self.retweets = retweets
        self.__refresh_twitter_values()
        self.s_twitter = requests.session()
        self.headers = {
            'x-guest-token': self.gt,
            'authorization': 'Bearer %s' % (self.bearer_token),
            'content-type': 'application/json'
        }
        self.user_id = self.__get_user_id()

    def __refresh_twitter_values(self):
        (
            self.gt,
            self.bearer_token,
            self.query_id
        ) = self.__get_new_twitter_values()

    def __get_new_twitter_values(self):
        '''
        Collect the values needed to make an unauthenticated request to the
        Twitter API for a user's timeline, using GetTwitterValues.

        Raises:
            FailedToGetTweetsException: if values could not be retrieved.

        Return:
            gt, bearer_token, query_id (tuple): the API values.
        '''

        twitter_values = GetTwitterValues()
        try:
            gt = twitter_values.get_guest_token()
            bearer_token = twitter_values.get_bearer_token()
            query_id = twitter_values.get_query_id()
        except FailedToGetTwitterValueException as e:
            raise FailedToGetTweetsException(e)
        return (gt, bearer_token, query_id)

    def get_twitter_values(self):
        '''
        Returns the current tokens retrieved from Twitter.

        Returns:
            tuple: the guest token, bearer token, and query id
        '''

        return (self.gt, self.bearer_token, self.query_id)

    def __get_user_id(self):
        '''
        Gets the id of the Twitter username supplied, by querying the GraphQL
        API's "UserByScreenName" operation.

        Raises:
            FailedToGetTweetsException: if the user id could not be retrieved.

        Returns:
            user_id (str): the user id.
        '''
        user_id = None
        url = (
            'https://api.twitter.com/graphql/%s/UserByScreenName'
            % (self.query_id)
        )
        params = {
            'variables': (
                '{"screen_name":"%s","withHighlightedLabel":true}'
                % (self.user)
            )
        }
        try:
            r = self.s_twitter.get(
                url,
                params=params,
                headers=self.headers
            )
            graph_ql_json = r.json()
        except Exception as e:
            raise FailedToGetTweetsException(
                'Failed to get the user id, request excepted with: %s'
                % (str(e))
            )
        try:
            user_id = graph_ql_json['data']['user']['rest_id']
        except KeyError:
            raise FailedToGetTweetsException(
                'Failed to get the user id, could not find user rest_id in '
                'GraphQL response.'
            )
        return user_id

    def get_tweets(
        self,
        count=None,
        refresh_tokens=False,
        last_id=None
    ):
        '''
        Queries the Twitter API using a guest token and authorization bearer
        token retrived from GetTwitterValues().

        Arguments:
            count (int): the amount of tweets to get.
            refresh_tokens (bool): get new tokens before checking for tweets.
            last_id (int): the id of the latest seen tweet.

        Raises:
            FailedToGetTweetsException: if tweets could not be retrieved.

        Returns:
            tweets (list): a list of tweet timestamp and text tuples for the
                user, sorted ascending in time, limited by the 'limit'
                argument.
        '''

        if refresh_tokens:
            self.__refresh_twitter_values()

        tweets = list()

        try:
            url = (
                'https://api.twitter.com/2/timeline/profile/%s.json'
                % self.user_id
            )
            # the 'count' param in this query is actually a maximum,
            # where deleted or suspended tweets are removed after the
            # count is applied, so we don't supply a count param in
            # the API query and instead apply it to the response data
            # later.
            params = {
                'tweet_mode': 'extended'
            }
            r = self.s_twitter.get(url, headers=self.headers, params=params)
            timeline_json = r.json()
        except Exception as e:
            raise FailedToGetTweetsException(
                'Failed to get tweets, request excepted with: %s'
                % (str(e))
            )
        try:
            tweets_json = timeline_json['globalObjects']['tweets']
            if self.retweets:
                tweet_ids = list(int(x) for x in tweets_json.keys())
            else:
                tweet_ids = list(
                    int(x) for x in tweets_json.keys()
                    if 'retweeted_status_id_str' not in tweets_json[x]
                )
            # if a last id value is supplied, only return tweets which an id
            # greater
            if last_id:
                tweet_ids = list(x for x in tweet_ids if x > last_id)
            if len(tweet_ids) > 0:
                # an assumption here is the display should be oldest to newest,
                # i.e. opposite to Twitter's UI, as it makes more sense in a
                # cli environment.
                tweet_ids.sort(reverse=True)
                tweet_ids_culled = tweet_ids[:count]
                tweet_ids_culled.sort()
                tweets = (
                    list(
                        (
                            tweets_json[str(x)]['created_at'],
                            unescape(tweets_json[str(x)]['full_text']),
                            tweets_json[str(x)]['id_str']
                        )
                        for x in tweet_ids_culled
                    )
                )
        except KeyError:
            raise FailedToGetTweetsException('Failed to get tweets.')
        return tweets



from bs4 import BeautifulSoup
from twittertail.exceptions import (
    FailedToGetMainJsException,
    FailedToGetTwitterValueException
)
import re
import requests


class GetTwitterValues:
    '''
    GetTwitterValues handles retrieving values from the Twitter markup that
    the frontend uses to query the Twitter API without authentication, such
    as the guest token and the authorization bearer token.
    '''

    def __init__(self):
        self.main_js = self.__get_main_js()

    def __get_main_js(self):
        '''
        Gets the contents of the 'main.XXXX.js' file that Twitter is currently
        using in its markup.

        Raises:
            FailedToGetMainJsException: if the main.XXXX.js file can't be
            retrived.

        Return:
            r.text (str): the contents of the main.XXXX.js file.
        '''
        main_js_href = ''
        try:
            r = requests.get('https://twitter.com')
        except Exception as e:
            raise FailedToGetMainJsException(
                'Failed to find the main.XXXX.js file, request excepted '
                'with: %s'
                % (str(e))
            )
        soup = BeautifulSoup(r.text, 'html.parser')
        for script in soup.find_all(
            href=re.compile(r'responsive-web/web_legacy/main\.[^\.]+\.js$')
        ):
            main_js_href = script['href']
        if not main_js_href or not main_js_href.startswith('https://'):
            raise FailedToGetMainJsException(
                'Failed to find the main.XXXX.js file'
            )
        try:
            r = requests.get(main_js_href)
        except Exception as e:
            raise FailedToGetMainJsException(
                'Failed to get the main.XXXX.js contents, request excepted '
                'with: %s'
                % (str(e))
            )
        return r.text

    def get_query_id(self):
        '''
        Extracts the GraphQL query id for the 'UserByScreenName' op from the
        Twitter markup. This is set in a JS file which is sourced into the
        main Twitter markup.

        Raises:
            FailedToGetTwitterValueException: if the user id can't be retrived

        Return:
            query_id (str): the query id
        '''

        query_id = ''
        query_id_pattern = re.compile(
            '.*queryId:"([^"]+)",operationName:"UserByScreenName",'
            'operationType:"query".*'
        )
        matches = re.match(query_id_pattern, self.main_js)
        if matches:
            query_id = matches.group(1)
        if not query_id:
            raise FailedToGetTwitterValueException(
                'Failed to get the user GraphQL query id from the Twitter '
                'markup.'
            )
        return query_id

    def get_bearer_token(self):
        '''
        Extracts the "bearer token" from the Twitter markup for use in
        unauthenticated API requests. This is set in a JS file which is
        sourced into the main Twitter markup.

        Raises:
            FailedToGetTwitterValueException: if the token can't be retrieved

        Returns:
            bearer_token (str): the bearer token value
        '''

        bearer_token = ''
        # we assume this pattern will apply to all future tokens.
        token_pattern = re.compile('.*s="(AAAA[^"]+).*')
        matches = re.match(token_pattern, self.main_js)
        if matches:
            bearer_token = matches.group(1)
        if not bearer_token:
            raise FailedToGetTwitterValueException(
                'Failed to get the bearer token from the Twitter markup.'
            )
        return bearer_token

    def get_guest_token(self):
        '''
        Extracts the "guest token" from the Twitter markup for use in
        unauthenticated API requests. This is set in an inline JS block.

        Raises:
            FailedToGetTwitterValueException: if the token can't be retrieved

        Returns:
            gt (str): the guest token value
        '''

        gt = ''
        try:
            r = requests.get('https://twitter.com')
        except Exception as e:
            raise FailedToGetTwitterValueException(
                'Failed to find the guest token, request excepted with: %s'
                % (str(e))
            )
        soup = BeautifulSoup(r.text, 'html.parser')
        for script in soup.find_all('script'):
            if 'gt=' in str(script):
                pattern = re.compile('.*gt=([^;]+);.*', re.DOTALL)
                matches = re.match(pattern, str(script))
                if matches:
                    gt = matches.groups(1)[0]
        if not gt:
            raise FailedToGetTwitterValueException(
                'Failed to find the guest token in Twitter markup.'
            )
        return gt



#!/usr/bin/python3
from setuptools import setup
setup(
    name='twittertail',
    packages=['twittertail'],
    version='0.3',
    license='MIT',
    description='Tail a twitter account from the command line.',
    author='Nathan Davison',
    author_email='ndavison85@gmail.com',
    url='https://github.com/ndavison/twittertail',
    download_url='https://github.com/ndavison/twittertail/archive/v03.tar.gz',
    keywords=['twitter', 'tweets', 'cli'],
    install_requires=[
        'requests',
        'beautifulsoup4',
        'asyncio',
        'aiohttp',
        'colored',
    ],
    entry_points={
        'console_scripts': [
            'twittertail=twittertail:main',
        ],
    },
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)
